#coding utf-8

require_relative 'dataset.rb'
require_relative 'learner.rb'

module MachineLearner
  # 複数の識別器から最良のものを自動選択する学習器
  class SimpleLearner < Learner
    
    # データを元に学習を行う
    # classifiersの各戻り値に対し識別を行い、結果のもっともよいものを採用する
    # @param datas [Array<DataSet>] トレーニングデータの配列
    # @param ds [Array<Float>] 各トレーナーデータの重みの配列
    # @return [Array<Boolean>] 識別結果の配列
    def learn(datas, ds)
      @log = []
      classifiers(datas, ds) do |cl|
        results = datas.map{|data| cl.test(data) }
        epsilon = results.zip(ds).map{|r, w| r ? 0 : w }.inject(:+)
        @log << {epsilon: epsilon, classifier: cl, results: results}
      end
      
      @best = @log.min {|h1, h2| h1[:epsilon] <=> h2[:epsilon] }
      @best[:results]
    end
    
    # 識別器を列挙する
    # 識別器はyieldでblockに渡す
    # @param datas [Array<DataSet>] トレーニングデータの配列
    # @param ds [Array<Float>] 各トレーナーデータの重みの配列
    def classifiers(datas, ds)
      raise NotImplementedError.new
    end
    
    # 識別器を文字列で表現する
    # @return [String] 識別器を表す文字列
    def to_s
      cl.to_s
    end
    
    # 識別を行う
    # @param x 特徴空間
    # @return [Fixnum] 識別結果
    def classify(x)
      cl.classify(x)
    end
    
    private
    
    # 性能の最も良い識別器を取得する
    # @return [Classifier] 性能の最も良い識別器を取得する
    def cl
      @best[:classifier]
    end
    
    # データを識別器に掛けた結果
    # ハッシュの配列であり、各ハッシュはキー[:classifier, :result, :epsiron]を持つ
    # @return log [Array<Hash>] 識別結果の配列
    def log
      @log
    end
    
  end
  
  # AdaBoost クラス
  class AdaBoost < Learner
    
    # コンストラクタ
    # @param learners [Array<Learner>] 学習器のリスト
    def initialize(learners)
      @learners = learners
      @alphas = []
    end
    
    # データを元に学習を行う
    # @param datas [Array<DataSet>] トレーニングデータの配列
    # @return [Array<Boolean>] 識別結果の配列
    def learn(datas, ds = nil)
      ds ||= [1.0 / datas.size] * datas.size
      
      @learners.each do |learner|
        # 学習記に順番に学習させる
        # results : トレーニングデータの正解、不正解の配列
        results = learner.learn(datas, ds)
        # epsilon : 学習器のエラー率(失敗した学習データの重みの総和)
        epsilon = results.zip(ds).map{|r, w| r ? 0 : w }.inject(:+)
        epsilon = ds.min * 0.1 if(epsilon == 0)
        # alpha : 学習器の重み(エラー率epsilonが低いほど高い値を取る)
        alpha = Math.log((1 - epsilon) / epsilon) / 2
        @alphas << alpha
        # 重みの更新
        ds = ds.map.with_index{|w, i| w * Math.exp(alpha * (results[i] ? -1 : 1)) }
        z = ds.inject(&:+)
        ds.map!{|w| w / z}
      end
    end
    
    # 識別結果の生の値を返す
    # @param x 特徴空間
    # @return [Float] 識別結果の生の値(-1..1)
    def classify_raw(x)
      [@learners, @alphas].transpose.reduce(0) { |score, (l, a)| score += l.classify(x) * a }
    end
    
    # 識別を行う
    # @param x 特徴空間
    # @return [Fixnum] 識別結果
    def classify(x)
      classify_raw(x) > 0 ? 1 : -1
    end
    
    # Learnerを表す文字列
    def to_s
      [@learners, @alphas].transpose.map {|l, a| "(#{a.round(3)} * #{l})" }.join(" + ");
    end
  end
end

# vim: set et ts=2 sts=2 sw=2:
